
#ifdef _WIN32
 #define _CRT_SECURE_NO_WARNINGS
#endif
#include <windows.h>
#include <wincrypt.h>

#include <cstdlib>
#include <cstdio>
#include <cstring>

#include <string>
#include <iostream>

// CSPs we work with
#define PRIVATE_CSP                 "EnterSafe ePass2003 CSP v1.0"
#define MICROSOFT_MINIDRIVER_CSP    "Microsoft Base Smart Card Crypto Provider"

// CSP to use
#define CSP_NAME PRIVATE_CSP

#define KEY_CONTAINER_NAME  "demo-key-container"


int main(int argc, char **argv)
{
    BOOL bResult = FALSE;

    HCRYPTPROV hProv = NULL;
    HCRYPTKEY hKey = NULL;
    HCRYPTHASH hHash = NULL;

    // First we aquire a CAPI context to the CSP

    bResult = CryptAcquireContext(&hProv, KEY_CONTAINER_NAME, CSP_NAME, PROV_RSA_FULL, 0);
    if (!bResult)
    {
        DWORD err = GetLastError();

        if (err == NTE_BAD_KEYSET)
        {
            printf("Demo key container does not exist. Creating it now...\n");

            bResult = CryptAcquireContext(&hProv, KEY_CONTAINER_NAME, CSP_NAME, PROV_RSA_FULL, CRYPT_NEWKEYSET);
            if (bResult)            
                printf("Success!\n");
            else
            {
                printf("CryptAcquireContext() failed to create the demo key container.\n");
                printf("Aborting.\n");
                return -1;
            }
        }
        else if (err == NTE_KEYSET_NOT_DEF)
        {
            printf("Error: The requested CSP is not installed.\n");
            return -1;
        }
        else
        {
            printf("CryptAcquireContext() failed (error %08x)\n", err);
            return -1;
        }
    }

    // Next we obtain the signing key from the container (create it if it doesn't exist yet)

    bResult = CryptGetUserKey(hProv, AT_SIGNATURE, &hKey);
    if (!bResult)
    {
        printf("No signing key present. Creating one now...\n");

        bResult = CryptGenKey(hProv, AT_SIGNATURE, 0, &hKey);
        if (bResult)        
            printf("Success!\n");        
        else
        {
            printf("CryptGenKey() failed (error %08x)\n", GetLastError());
            printf("Aborting.\n");
            return -1;
        }
    }    

    // Now we can sign some data!    

    // Ask which file to calculate signature of
    std::string filename = "";
    std::cout << "Which file do you want to calculate the signature for? " << std::endl;
    std::getline(std::cin, filename);

    // Open the file
    FILE *f = fopen(filename.c_str(), "rb");
    if (!f)
    {
        printf("Failed to open \"%s\"\n", filename.c_str());
        printf("Aborting.\n");
        return -1;
    }

    // Create a Hash object
    bResult = CryptCreateHash(hProv, CALG_SHA_256, 0, 0, &hHash);
    if (!bResult)
    {
        printf("CryptCreateHash() failed (error %08x)\n", GetLastError());
        printf("Aborting.\n");
        return -1;
    }

    // Read the file in chunks, adding each chunk to the hash object
    size_t dataLen = 0;
    size_t dataTotal = 0;
    unsigned char buffer[2048];
    do
    {
        dataLen = fread(buffer, 1, sizeof(buffer), f);
        if (dataLen)
        {
            dataTotal += dataLen;

            bResult = CryptHashData(hHash, buffer, dataLen, 0);
            if (!bResult)
            {
                printf("CryptHashData() failed (error %08x)\n", GetLastError());
                printf("Aborting.\n");
                return -1;
            }
        }
    } while (dataLen > 0);

    fclose(f);

    printf("Added %u bytes to digest", dataTotal);

    // Request the size of the signature so we can allocate a buffer
    DWORD sigLen = 0;
    bResult = CryptSignHash(hHash, AT_SIGNATURE, NULL, 0, NULL, &sigLen);
    if (!bResult)
    {
        printf("CryptSignHash() failed (error %08x) when getting size of signature\n", GetLastError());
        printf("Aborting.\n");
        return -1;
    }

    // Allocate a buffer for the signature
    unsigned char *sigBuffer = (unsigned char *) malloc(sigLen);
    if (!sigBuffer)
    {
        printf("Failed to allocate %u bytes for signature buffer\n", sigLen);
        printf("Aborting.\n");
        return -1;
    }

    // Now get the actual signature
    bResult = CryptSignHash(hHash, AT_SIGNATURE, NULL, 0, sigBuffer, &sigLen);
    if (!bResult)
    {
        printf("CryptSignHash() failed (error %08x) when creating the signature\n", GetLastError());
        printf("Aborting.\n");
        return -1;
    }

    printf("Signature is:\n");
    for (unsigned int i=0; i<sigLen; i++)
        printf("%02x ", sigBuffer[i]);
    printf("\n");

    // Note: CryptVerifySignature() can be used to verify the signature of the data using the public key.

    // Cleanup
    CryptDestroyHash(hHash);
    CryptReleaseContext(hProv, 0);

    return 0;
}

