
/*
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

package com.microcosm.pki;


import java.io.*;
import java.util.*;

import java.io.StringWriter;

import java.security.KeyStore;
import java.security.KeyPairGenerator;
import java.security.KeyPair;
import java.security.KeyFactory;
import java.security.PublicKey;
import java.security.PrivateKey;
import java.security.SecureRandom;
import java.security.spec.*;
import java.security.cert.X509Certificate;
import java.security.Signature;
import java.math.BigInteger;

import java.security.NoSuchAlgorithmException;
import java.security.KeyStoreException;
import java.security.cert.CertificateException;
import java.security.InvalidAlgorithmParameterException;
import java.security.InvalidKeyException;
import java.security.NoSuchProviderException;
import java.security.SignatureException;
import java.security.UnrecoverableKeyException;
import java.security.cert.Certificate;
import java.security.cert.CertificateParsingException;

import org.bouncycastle.asn1.DERBMPString;
import org.bouncycastle.asn1.pkcs.PKCSObjectIdentifiers;
import org.bouncycastle.asn1.x500.X500Name;
import org.bouncycastle.asn1.x500.X500NameBuilder;
import org.bouncycastle.asn1.x500.style.BCStyle;
import org.bouncycastle.asn1.x509.BasicConstraints;
import org.bouncycastle.asn1.x509.Extension;
import org.bouncycastle.cert.X509CertificateHolder;
import org.bouncycastle.cert.X509v1CertificateBuilder;
import org.bouncycastle.cert.X509v3CertificateBuilder;
import org.bouncycastle.cert.jcajce.JcaX509CertificateConverter;
import org.bouncycastle.cert.jcajce.JcaX509ExtensionUtils;
import org.bouncycastle.cert.jcajce.JcaX509v1CertificateBuilder;
import org.bouncycastle.cert.jcajce.JcaX509v3CertificateBuilder;
import org.bouncycastle.jce.interfaces.PKCS12BagAttributeCarrier;
import org.bouncycastle.jce.provider.BouncyCastleProvider;
import org.bouncycastle.operator.ContentSigner;
import org.bouncycastle.operator.jcajce.JcaContentSignerBuilder;
import org.bouncycastle.operator.OperatorCreationException;
import org.bouncycastle.openssl.jcajce.JcaPEMWriter;
import org.bouncycastle.asn1.ASN1InputStream;
import org.bouncycastle.asn1.ASN1Primitive;
import org.bouncycastle.asn1.ASN1Sequence;
import org.bouncycastle.jce.provider.X509CertificateObject;
import org.bouncycastle.util.io.pem.PemObject;
import org.bouncycastle.util.io.pem.PemReader;

import iaik.pkcs.pkcs11.Mechanism;
import iaik.pkcs.pkcs11.Module;
import iaik.pkcs.pkcs11.Session;
import iaik.pkcs.pkcs11.Slot;
import iaik.pkcs.pkcs11.Token;
import iaik.pkcs.pkcs11.TokenException;
import iaik.pkcs.pkcs11.objects.AESSecretKey;
import iaik.pkcs.pkcs11.objects.DES3SecretKey;
import iaik.pkcs.pkcs11.objects.DESSecretKey;
import iaik.pkcs.pkcs11.wrapper.PKCS11Constants;
import iaik.pkcs.pkcs11.wrapper.*;
import iaik.pkcs.pkcs11.objects.Object;
import iaik.pkcs.pkcs11.objects.RC4SecretKey;
import iaik.pkcs.pkcs11.objects.RSAPublicKey;
import iaik.pkcs.pkcs11.objects.X509PublicKeyCertificate;
import iaik.pkcs.pkcs11.parameters.InitializationVectorParameters;
import iaik.pkcs.pkcs11.DefaultInitializeArgs;

public class PKISample
{
    private static final String LIBRARY_NAME = "mpki";
    private static final String FILE_SEPARATOR = System.getProperty("file.separator");
    //private static final String USER_PIN = "12345678";     // IMPORTANT: This would not be hard coded in production code
    private static final String KEY_ALIAS = "com.microcosm.pki.PKISample.key";
    private static int KEY_SIZE_BITS = 2048;

    // Instance members
    private java.security.Provider pkcs11Provider;
    private KeyStore keyStore;
    private BouncyCastleProvider bcProvider;

    PKISample()
    {

    }

    private void init()
    {
        this.pkcs11Provider = createPKCS11Provider();
        this.bcProvider = new BouncyCastleProvider();
    }

    private static java.security.Provider createPKCS11Provider()
    {
        String cwd = System.getProperty("user.dir");

        String libFilename = LIBRARY_NAME + PlatformUtil.jvmBitness() + PlatformUtil.librarySuffix();
        System.out.println("Using: " + libFilename);

        // Define parameters for Sun PKCS#11 provider
        String configStr =
            "name = Microcosm-USB-PKI-Token\n" +
            "library = " + cwd + FILE_SEPARATOR + libFilename + "\n" +
            "slot = 1 \n" +
            "attributes = compatibility \n" +
            "attributes(*,*,*)=\n{\nCKA_TOKEN=true\nCKA_LOCAL=true\n}\n" +
            "attributes(*,CKO_PUBLIC_KEY,*)=\n{\nCKA_ENCRYPT=true\nCKA_VERIFY=true\nCKA_VERIFY_RECOVER=true\nCKA_WRAP=true\n}\n" +
            "attributes(*,CKO_PRIVATE_KEY,*)=\n{\nCKA_PRIVATE=true\nCKA_DECRYPT=true\nCKA_SIGN=true\nCKA_UNWRAP=true\nCKA_SENSITIVE=true\nCKA_EXTRACTABLE=false\n}\n";

        return new sun.security.pkcs11.SunPKCS11(new ByteArrayInputStream(configStr.getBytes()));
    }

    private KeyStore login()
        throws IOException,
        KeyStoreException,
        NoSuchAlgorithmException,
        CertificateException
    {
        // Acquire User PIN from stdin
        BufferedReader input = new BufferedReader(new InputStreamReader(System.in));
        System.out.print("Please give the user PIN for the token: ");
        String userPIN = input.readLine();
        char[] pin = userPIN.toCharArray();

        //Initialize Key Store logging in using the PIN provided
        KeyStore keyStore = KeyStore.getInstance("PKCS11", this.pkcs11Provider);
        keyStore.load(null, pin);
        return keyStore;
    }

    private void initAndLogin()
        throws IOException,
        KeyStoreException,
        NoSuchAlgorithmException,
        CertificateException
    {
        // Initialize providers etc
        init();

        // Login to the token to aquire a KeyStore instance
        this.keyStore = login();
    }

    private void generateKey()
        throws NoSuchAlgorithmException,
        InvalidAlgorithmParameterException,
        OperatorCreationException,
        CertificateException,
        InvalidKeyException,
        NoSuchProviderException,
        KeyStoreException,
        SignatureException
    {
        // ==================
        // Create the keypair
        // ==================
        KeyPairGenerator keygen = KeyPairGenerator.getInstance("RSA", this.pkcs11Provider);
        AlgorithmParameterSpec spec = new RSAKeyGenParameterSpec(KEY_SIZE_BITS, RSAKeyGenParameterSpec.F4);
        keygen.initialize(spec);
        KeyPair kp = keygen.generateKeyPair();
        PublicKey publicKey = kp.getPublic();
        PrivateKey privateKey = kp.getPrivate();

        // =========================================================================
        // Build the certificate that we will use to add the keypair to the KeyStore
        // =========================================================================
        // In this example we use the same value for issuer as the alias we use for the KeyStore
        X500Name issuer = new X500Name("CN=" + KEY_ALIAS);
        // Same issuer and owner (since this is a self-signed certificate)
        X500Name owner = issuer;
        // Certificate Not Valid Before: now
        Date notBefore = new Date(System.currentTimeMillis());
        // Certificate Not Valid After: 1 year from now
        Calendar cal = Calendar.getInstance();
        cal.setTime(notBefore);
        cal.add(Calendar.YEAR, 1);
        Date notAfter = cal.getTime();
        // Build the certificate
        X509v3CertificateBuilder builder = new JcaX509v3CertificateBuilder(
            issuer,                                         // Certificate Issuer
            new BigInteger(64, new SecureRandom()),         // Serial Number (random)
            notBefore,                                      // Valid From time
            notAfter,                                       // Valid Until time
            owner,                                          // Certificate Owner (the Subject)
            publicKey                                       // Public Key we create above
        );
        // Sign the certificate
        ContentSigner signer = new JcaContentSignerBuilder("SHA256WithRSA").setProvider(this.pkcs11Provider).build(privateKey);
        X509CertificateHolder certHolder = builder.build(signer);
        X509Certificate cert = new JcaX509CertificateConverter().setProvider(this.bcProvider).getCertificate(certHolder);

        // Test verify the certificate we just created!
        cert.verify(publicKey);

        this.keyStore.setKeyEntry(KEY_ALIAS, privateKey, new char[0], new java.security.cert.Certificate[] {cert});

        return;
    }

    private void signDataExample()
        throws KeyStoreException,
        NoSuchAlgorithmException,
        UnrecoverableKeyException,
        InvalidKeyException,
        SignatureException,
        IOException
    {
        // First we ask what file they want to calculate the digital signature for
        BufferedReader in = new BufferedReader(new InputStreamReader(System.in));
        FileInputStream dataFileStream = null;
        String filename = "";
        do
        {
            dataFileStream = null;

            System.out.println("File name to calculate digital signature of:");
            try
            {
                filename = in.readLine();
            }
            catch (IOException ioe)
            {
                continue;
            }

            try
            {
                dataFileStream = new FileInputStream(filename);
            }
            catch (IOException ioe)
            {
                System.out.println("Failed to open file \""+filename+"\".\nException:" + ioe.toString());
                ioe.printStackTrace();
                System.out.println();
            }
        } while (dataFileStream == null);

        System.out.println("Successfully opened \""+filename+"\"");
        System.out.println("\n");

        // Ask which certificate to use for data signing operation
        Enumeration<String> _aliases = this.keyStore.aliases();
        List aliases = Collections.list(_aliases);
        System.out.println("Available Certificate Aliases:");
        for (java.lang.Object _alias : aliases)
        {
            String alias = (String)_alias;
            // Can only sign with this alias if there is a PrivateKey associated with the alias
            if (this.keyStore.getKey(alias, new char[0]) != null)
                System.out.println(alias);
        }
        System.out.print("Type the alias you wish to use: ");
        String aliasToSignWith = in.readLine();

        // Now we calculate the digital signature of the data in the file.
        // The steps are:
        // 1. Obtain a PrivateKey object by fetching the key by alias
        // 2. Create a signature for the data by reading the input file in chunks, then finalizing the signature

        // 1
        PrivateKey privateKey = (PrivateKey) this.keyStore.getKey(aliasToSignWith, new char[0]);

        // 2
        // Create and initialize a Signature object
        Signature sig = Signature.getInstance("SHA256WithRSA", this.pkcs11Provider);
        sig.initSign(privateKey);
        // Read the input file, adding the data to the Signature object
        int bytesRead = 0;
        int totalBytesRead = 0;
        do
        {
            bytesRead = 0;
            byte[] buffer = new byte[1024];

            try
            {
                bytesRead = dataFileStream.read(buffer, 0, buffer.length);
            }
            catch (IOException ioe)
            {
                System.out.println("IOException reading input file: " + ioe.toString());
                ioe.printStackTrace();
            }

            if (bytesRead > 0)
            {
                sig.update(buffer, 0, bytesRead);
                totalBytesRead += bytesRead;
            }
        } while (bytesRead != -1);

        System.out.println("Read " + totalBytesRead + " bytes from input file.");

        // Create the signature
        byte[] sigBytes = sig.sign();

        // Display the signature
        if ((sigBytes != null) && (sigBytes.length > 0))
        {
            System.out.println("");
            for (int i=0; i<sigBytes.length; i++)
                System.out.print(new Formatter().format("%02x ", sigBytes[i]).toString());
        }
        System.out.println("");

        return;
    }

    private void displayCertificatePEM()
        throws KeyStoreException,
        IOException
    {
        System.out.println("Certificate is:");
        Certificate cert = this.keyStore.getCertificate(KEY_ALIAS);
        StringWriter sw = new StringWriter();
        JcaPEMWriter pemWriter = new JcaPEMWriter(sw);
        pemWriter.writeObject(cert);
        pemWriter.close();
        System.out.println(sw.toString());
    }

    public static boolean importCert()
    {
        try
        {
            Module module_ = Module.getInstance("mpki"+PlatformUtil.jvmBitness()+".dll", "iaik-pkcs11-wrapper"+PlatformUtil.jvmBitness()+".dll");
            module_.initialize(new DefaultInitializeArgs());
            Slot[] slots_ = module_.getSlotList(Module.SlotRequirement.TOKEN_PRESENT);
            if (slots_.length == 0)
            {
                System.out.println("No slot with present token found!");
                return false;
            }
            Token token_ = slots_[0].getToken();

            Session session_;
            // Open Session and Login
            try
            {
                session_ = token_.openSession(Token.SessionType.SERIAL_SESSION,
                Token.SessionReadWriteBehavior.RW_SESSION, null, null);
                System.out.println("Please input user pin:");
                Scanner scanner = new Scanner(System.in);
                String strUserPin = scanner.next();
                char[] usrPIN = strUserPin.toCharArray();
                session_.login(Session.UserType.USER, usrPIN);
            }
            catch (PKCS11Exception e)
            {
                if (e.getErrorCode() == PKCS11Constants.CKR_USER_ALREADY_LOGGED_IN)
                {
                    System.out.println("User already logged in");
                }
                e.printStackTrace();
                return false;
            }
            catch ( TokenException e )
            {
                e.printStackTrace();
                return false;
            }

            BufferedReader stdinReader = new BufferedReader(new InputStreamReader(System.in));

            // Import P12 file
            System.out.println("Please give the filename of the PFX file:");
            String pfxFilename = stdinReader.readLine();
            InputStream pfxStream = new FileInputStream(new File(pfxFilename));
            // Get the password for the P12 file
            System.out.println("Please give the PFX password:");
            String pfxPassword = stdinReader.readLine();

            // Load a KeyStore from the P12 file
            KeyStore keystore = KeyStore.getInstance("PKCS12");
            keystore.load(pfxStream, pfxPassword.toCharArray());

            pfxStream.close();

            // Display all key aliases available
            Enumeration<String> aliases = keystore.aliases();
            ArrayList<String> itr = java.util.Collections.list(aliases);
            System.out.println("Available Key Aliases:");
            for (String s : itr)
            {
                System.out.println(s);
            }
            System.out.println("Which key are you looking to import:");
            String importAlias = stdinReader.readLine();
            PrivateKey key = (PrivateKey) keystore.getKey(importAlias, pfxPassword.toCharArray());
            X509Certificate x509Cert = (X509Certificate) keystore.getCertificate(importAlias);

            PublicKey pubKey = (PublicKey)x509Cert.getPublicKey();
            PrivateKey prikey = (PrivateKey)keystore.getKey(importAlias, pfxPassword.toCharArray());

            // Now we can create the required objects...

            // Build the Public Key object
            iaik.pkcs.pkcs11.objects.RSAPublicKey pub = new iaik.pkcs.pkcs11.objects.RSAPublicKey();
            pub.getToken().setBooleanValue(Boolean.TRUE);
            pub.getSubject().setByteArrayValue(x509Cert.getSubjectDN().getName().getBytes());
            pub.getWrap().setBooleanValue(Boolean.TRUE);
            pub.getEncrypt().setBooleanValue(Boolean.TRUE);
            pub.getVerify().setBooleanValue(Boolean.TRUE);
            pub.getVerifyRecover().setBooleanValue(Boolean.TRUE);
            java.security.interfaces.RSAPublicKey  pubb = ( java.security.interfaces.RSAPublicKey )pubKey;
            byte[] modulus = iaik.pkcs.pkcs11.Util.unsignedBigIntergerToByteArray(pubb.getModulus());
            pub.getModulus().setByteArrayValue(modulus);
            pub.getPublicExponent().setByteArrayValue(pubb.getPublicExponent().toByteArray());
            // Save the Public Key object to the token
            session_.createObject(pub);

            // Build the Private Key object
            java.security.interfaces.RSAPrivateKey  priO = ( java.security.interfaces.RSAPrivateKey )prikey;
            iaik.pkcs.pkcs11.objects.RSAPrivateKey pri = new iaik.pkcs.pkcs11.objects.RSAPrivateKey();
            pri.getToken().setBooleanValue(Boolean.TRUE);
            pri.getSubject().setByteArrayValue(x509Cert.getSubjectDN().getName().getBytes());
            byte[] modulusPri = iaik.pkcs.pkcs11.Util.unsignedBigIntergerToByteArray(priO.getModulus());
            pri.getModulus().setByteArrayValue(modulusPri);
            pri.getPrivateExponent().setByteArrayValue(priO.getPrivateExponent().toByteArray());
            pri.getPublicExponent().setByteArrayValue(pubb.getPublicExponent().toByteArray());
            pri.getUnwrap().setBooleanValue(Boolean.TRUE);
            // Save the Private Key object to the token
            session_.createObject(pri);

            // Build the X509 certificate object
            iaik.pkcs.pkcs11.objects.X509PublicKeyCertificate pkcs11X509PublicKeyCertificate = new iaik.pkcs.pkcs11.objects.X509PublicKeyCertificate();
            pkcs11X509PublicKeyCertificate.getToken().setBooleanValue(Boolean.TRUE);
            pkcs11X509PublicKeyCertificate.getSubject().setByteArrayValue(x509Cert.getSubjectDN().getName().getBytes());
            pkcs11X509PublicKeyCertificate.getIssuer().setByteArrayValue(x509Cert.getIssuerDN().getName().getBytes() );
            pkcs11X509PublicKeyCertificate.getSerialNumber().setByteArrayValue(x509Cert.getSerialNumber().toByteArray());
            pkcs11X509PublicKeyCertificate.getValue().setByteArrayValue(x509Cert.getEncoded());
            // Save the certificate object to the token
            session_.createObject(pkcs11X509PublicKeyCertificate);

        }
        catch (Exception e)
        {
            e.printStackTrace();
        }
        return true;
    }

    public void run()
    {
        try
        {
            System.out.println("What would you like to do?");
            System.out.println("1  Sign some data");
            System.out.println("2  Import a key from a file");
            //System.out.println("q  Quit");
            System.out.print("Please select an option (hit Enter to quit): ");
            BufferedReader input = new BufferedReader(new InputStreamReader(System.in));
            String choice = input.readLine().trim();
            switch (choice)
            {
                case "1":
                    initAndLogin();

                    // Check if the demo key-pair exists
                    if (!this.keyStore.containsAlias(KEY_ALIAS))
                    {
                        // Key-pair doesn't exist yet...
                        // Generate our RSA key-pair now
                        System.out.println("Creating demo key-pair...");
                        generateKey();
                        System.out.println("Demo key-pair created OK.");
                    }
                    else
                    {
                        System.out.println("Demo key-pair already exists in token.");
                    }

                    // Show the certificate in PEM format
                    displayCertificatePEM();

                    // Sign some data
                    signDataExample();
                    break;

                case "2":
                    importCert();
                    break;

                case "q":
                default:
                    break;
            }
        }
        catch (Exception e)
        {
            e.printStackTrace();
        }

        return;
    }

    public static void main(String args[])
    {
        new PKISample().run();
    }

}

class PlatformUtil
{
    public static String jvmBitness()
    {
        if (System.getProperty("os.arch").contains("64"))
            return "64";
        return "32";
    }

    public static String librarySuffix()
    {
        String osName = System.getProperty("os.name");
        if (osName.contains("Windows"))
             return ".dll";
        else if (osName.contains("Mac"))
            return ".dylib";
        return ".so";
    }
}
